//
//  ViewController.m
//  VASTest
//
//  Created by Flex on 6/14/17.
//  Copyright © 2017 Datecs. All rights reserved.
//

#import "ViewController.h"
#import "DTDevices.h"
#import "Config.h"
#import "EMVTLV.h"
#import "EMVTags.h"
#import "EMVPrivateTags.h"
#import "ProgressViewController.h"
#import "EMVProcessorHelper.h"
#import <CommonCrypto/CommonDigest.h>

@interface ViewController ()
{
    IBOutlet UITextField *tfMID1;
    IBOutlet UITextField *tfMID2;
    IBOutlet UITextField *tfMID3;
    IBOutlet UITextField *tfMURL;

    IBOutlet UISegmentedControl *segProtocolMode;
    IBOutlet UISegmentedControl *segTerminalMode;

    DTDevices *dtdev;
}

@end

@implementation ViewController

void displayAlert(NSString *title, NSString *message)
{
    UIAlertView *alert = [[UIAlertView alloc] initWithTitle:title message:message delegate:nil cancelButtonTitle:@"Ok" otherButtonTitles:nil, nil];
    [alert show];
}

-(BOOL)textFieldShouldEndEditing:(UITextField *)theTextField;
{
    return YES;
}

- (BOOL)textFieldShouldReturn:(UITextField *)textField;
{
    [textField resignFirstResponder];
    [self savePrefs];
    return YES;
}

-(BOOL)textField:(UITextField *)textField shouldChangeCharactersInRange:(NSRange)range replacementString:(NSString *)string
{
    //limit the size to 32
    int limit = 32;
    return !([textField.text length]>=limit && [string length] > range.length);
}

NSData *vasGetMID(NSString *url)
{
    if(url.length==0)
        return nil;

    uint8_t hash[CC_SHA256_DIGEST_LENGTH];

    CC_SHA256(url.UTF8String,(int)url.length,hash);

    return [NSData dataWithBytes:hash length:sizeof(hash)];
}

-(void)printTagTree:(TLV *)tag receipt:(NSMutableString *)receipt ident:(int)ident
{
    for(int i=0;i<ident;i++)
        [receipt appendString:@" "];
    if(tag.constructed)
    {
        [receipt appendFormat:@"%X:\n",tag.tag];
        for (TLV *stag in [TLV decodeTags:tag.data]) {
            [self printTagTree:stag receipt:receipt ident:ident+1];
        }
    }else
    {
        [receipt appendFormat:@"%@\n",tag];
    }
}

//parse some tags in order to build a display receipt to show
-(NSMutableString *)buildDisplayReceipt:(NSArray *)tags
{
    //build a receipt to display
    NSMutableString *receipt=[NSMutableString string];
    NSLog(@"Tags: %@",tags);

    [receipt appendFormat:@"* Datecs Ltd *\n"];
    [receipt appendFormat:@"\n"];


    [receipt appendFormat:@"Terminal ID: %@\n",[EMVProcessorHelper decodeNib:[TLV findLastTag:TAG_9F1C_TERMINAL_ID tags:tags].data]];
    [receipt appendFormat:@"\n"];

    [receipt appendFormat:@"Date: %@ %@\n",
     [EMVProcessorHelper decodeDateString:[TLV findLastTag:TAG_9A_TRANSACTION_DATE tags:tags].data],
     [EMVProcessorHelper decodeTimeString:[TLV findLastTag:TAG_9F21_TRANSACTION_TIME tags:tags].data]
     ];

    [receipt appendFormat:@"* Payment *\n"];

    [receipt appendFormat:@"Transaction Result:\n"];
    int transactionResult=[EMVProcessorHelper decodeInt:[TLV findLastTag:TAG_C1_TRANSACTION_RESULT tags:tags].data];
    NSString *transactionResultString=nil;
    switch (transactionResult)
    {
        case EMV_RESULT_APPROVED:
            transactionResultString=@"APPROVED";
            break;
        case EMV_RESULT_DECLINED:
            transactionResultString=@"DECLINED";
            break;
        case EMV_RESULT_TRY_ANOTHER_INTERFACE:
            transactionResultString=@"TRY ANOTHER INTERFACE";
            break;
        case EMV_RESULT_TRY_AGAIN:
            transactionResultString=@"TRY AGAIN";
            break;
        case EMV_RESULT_END_APPLICATION:
            transactionResultString=@"END APPLICATION";
            break;
        case EMV_RESULT_LOYALTY_COMPLTE:
            transactionResultString=@"LOYALTY COMPLETE";
            break;
    }
    [receipt appendString:transactionResultString];
    [receipt appendFormat:@"\n"];

    TLV *t;
    int interface=EMV_INTERFACE_CONTACT;
    t=[TLV findLastTag:TAG_C3_TRANSACTION_INTERFACE tags:tags];
    if(t)
    {
        const uint8_t *bytes=t.data.bytes;
        interface=bytes[0];
        switch (interface) {
            case EMV_INTERFACE_CONTACT:
                [receipt appendString:@"Interface: contact\n"];
                break;
            case EMV_INTERFACE_CONTACTLESS:
                [receipt appendString:@"Interface: contactless\n"];
                break;
            case EMV_INTERFACE_MAGNETIC:
                [receipt appendString:@"Interface: magnetic\n"];
                break;
            case EMV_INTERFACE_MAGNETIC_MANUAL:
                [receipt appendString:@"Interface: manual entry\n"];
                break;
        }
    }

    for (TLV *tag in tags) {
        if(tag.tag==0xEB)
        {
            [receipt appendFormat:@"* APPLE VAS *\n"];
            int ident=0;
            NSArray<TLV *> *stags=[TLV decodeTags:tag.data];
            for(TLV *stag in stags)
            {
                [self printTagTree:stag receipt:receipt ident:ident];
            }
        }
    }


    //find and get Track1 masked and Track2 masked tags for display purposes
    NSString *t1Masked=nil;
    NSString *t2Masked=nil;

    t=[TLV findLastTag:TAG_D3_TRACK1_MASKED tags:tags];
    if(t)
        t1Masked=[[NSString alloc] initWithData:t.data encoding:NSASCIIStringEncoding];
    t=[TLV findLastTag:TAG_D4_TRACK2_MASKED tags:tags];
    if(t)
        t2Masked=[[NSString alloc] initWithData:t.data encoding:NSASCIIStringEncoding];

    return receipt;
}

static bool shouldUpdate=false;
-(void)updateScreen
{
    if(!shouldUpdate)
        return;
    //kill the emv engine
    [dtdev emv2Deinitialise:nil];
    //set the screen to default state
    [dtdev uiShowInitScreen:nil];
}

//called when emv transaction is finished
//data parameter contains all tags, sent and received during the transaction
//if you want specific tags to be received here too, add them in configuration tag E1 subtag C2
//otherwise you can just query for them
-(void)emv2OnTransactionFinished:(NSData *)data;
{
    [progressViewController.view removeFromSuperview];

    NSLog(@"emv2OnTransactionFinished: %@",data);

    shouldUpdate=true;
    [self performSelector:@selector(updateScreen) withObject:nil afterDelay:2.0];
    if(!data)
    {
        [dtdev emv2Deinitialise:nil];
        displayAlert(@"Error",@"Transaction could not be completed!");
        return;
    }


    NSArray *tags=[TLV decodeTags:data];

    int transactionResult=[EMVProcessorHelper decodeInt:[TLV findLastTag:TAG_C1_TRANSACTION_RESULT tags:tags].data];

    TLV *t;

    if(transactionResult==EMV_RESULT_APPROVED || transactionResult==EMV_RESULT_LOYALTY_COMPLTE)
    {
        //build a display receipt
        NSMutableString *receipt=[self buildDisplayReceipt:tags];

        displayAlert(@"Transaction complete!",receipt);

    }else
    {
        NSString *transactionResultString=nil;
        switch (transactionResult)
        {
            case EMV_RESULT_APPROVED:
                transactionResultString=@"APPROVED";
                break;
            case EMV_RESULT_DECLINED:
                transactionResultString=@"DECLINED";
                break;
            case EMV_RESULT_TRY_ANOTHER_INTERFACE:
                transactionResultString=@"TRY ANOTHER INTERFACE";
                break;
            case EMV_RESULT_TRY_AGAIN:
                transactionResultString=@"TRY AGAIN";
                break;
            case EMV_RESULT_END_APPLICATION:
                transactionResultString=@"END APPLICATION";
                break;
            case EMV_RESULT_LOYALTY_COMPLTE:
                transactionResultString=@"LOYALTY COMPLETE";
                break;
        }

        NSString *reasonMessage=@"Terminal declined";
        t=[TLV findLastTag:TAG_C4_TRANSACTION_FAILED_REASON tags:tags];
        if(t)
        {
            const uint8_t *bytes=t.data.bytes;
            int reason=bytes[0];
            if(reason==REASON_CANCELED)
                reasonMessage=@"User cancelled";
            if(reason==REASON_TIMEOUT)
                reasonMessage=@"Transaction timed out";
        }
        displayAlert(@"Transaction failed!",[NSString stringWithFormat:@"Transaction result: %@\nFailure reason: %@",transactionResultString,reasonMessage]);
    }
}

//in data buffer all the tags from configuration E5/C1 will be returned here (if available)
-(void)emv2OnOnlineProcessing:(NSData *)data;
{
    //for the demo fake a successful server response (30 30)
    NSData *serverResponse=[TLV encodeTags:@[[TLV tlvWithHexString:@"30 30" tag:TAG_8A_AUTH_RESP_CODE]]];
    NSData *response=[TLV encodeTags:@[[TLV tlvWithHexString:@"01" tag:0xC2],[TLV tlvWithData:serverResponse tag:0xE6]]];
    [dtdev emv2SetOnlineResult:response error:nil];
}


//notification from the pinpad about transaction message, you can use it to display something on the iOS screen
-(void)emv2OnUserInterfaceCode:(int)code status:(int)status holdTime:(NSTimeInterval)holdTime;
{
    NSString *ui=@"";
    NSString *uistatus=@"not provided";
    switch (code)
    {
        case EMV_UI_NOT_WORKING:
            ui = @"Not working";
            break;
        case EMV_UI_APPROVED:
            ui = @"Transaction Approved";
            break;
        case EMV_UI_DECLINED:
            ui = @"Transaction Declined";
            break;
        case EMV_UI_PLEASE_ENTER_PIN:
            ui = @"Please enter PIN";
            break;
        case EMV_UI_ERROR_PROCESSING:
            ui = @"Error processing";
            break;
        case EMV_UI_REMOVE_CARD:
            ui = @"Please remove card";
            break;
        case EMV_UI_IDLE:
            ui = @"Idle";
            break;
        case EMV_UI_PRESENT_CARD:
            ui = @"Please present card";
            break;
        case EMV_UI_PROCESSING:
            ui = @"Processing...";
            break;
        case EMV_UI_CARD_READ_OK_REMOVE:
            ui = @"It is okay to remove card";
            break;
        case EMV_UI_TRY_OTHER_INTERFACE:
            ui = @"Try another interface";
            break;
        case EMV_UI_CARD_COLLISION:
            ui = @"Card collision";
            break;
        case EMV_UI_SIGN_APPROVED:
            ui = @"Signature approved";
            break;
        case EMV_UI_ONLINE_AUTHORISATION:
            ui = @"Online authorization";
            break;
        case EMV_UI_TRY_OTHER_CARD:
            ui = @"Try another card";
            break;
        case EMV_UI_INSERT_CARD:
            ui = @"Please insert card";
            break;
        case EMV_UI_CLEAR_DISPLAY:
            ui = @"Clear display";
            break;
        case EMV_UI_SEE_PHONE:
            ui = @"See phone";
            break;
        case EMV_UI_PRESENT_CARD_AGAIN:
            ui = @"Please present card again";
            break;
        case EMV_UI_SELECT_APPLICAITON:
            ui = @"Select application on device";
            break;
        case EMV_UI_MANUAL_ENTRY:
            ui = @"Enter card on device";
            break;
        case EMV_UI_NA:
            ui = @"N/A";
            break;
    }
    switch (status)
    {
        case EMV_UI_STATUS_NOT_READY:
            uistatus = @"Status Not Ready";
            break;
        case EMV_UI_STATUS_IDLE:
            uistatus = @"Status Idle";
            break;
        case EMV_UI_STATUS_READY_TO_READ:
            uistatus = @"Status Ready To Read";
            break;
        case EMV_UI_STATUS_PROCESSING:
            uistatus = @"Status Processing";
            break;
        case EMV_UI_STATUS_CARD_READ_SUCCESS:
            uistatus = @"Status Card Read Success";
            break;
        case EMV_UI_STATUS_ERROR_PROCESSING:
            uistatus = @"Status Processing";
            break;
    }
    [progressViewController updateMessage:ui];
}

#define EMV_COMMAND(operation,c) {if(!c){displayAlert(@"Operation failed!",[NSString stringWithFormat:@"%@ failed, error %@, code: %d",operation,error.localizedDescription,(int)error.code]); [dtdev emv2Deinitialise:&error]; [progressViewController.view removeFromSuperview]; return;} }

-(void)emv2OnAppleVASProcessedWithResult:(int)result data:(NSData *)data;
{
    //parse apps, multiples of 6F templates
    NSArray<TLV *> *vas=[TLV decodeTags:data];

    NSMutableString *s=[NSMutableString string];

    NSString *vasResult=[NSString stringWithFormat:@"%04X",result];
    switch (result) {
        case 0x0000:
            vasResult=@"End Application";
            break;

        case 0x8000:
            vasResult=@"See Phone";
            break;

        case 0x1000:
            vasResult=@"Try Again";
            break;

        case 0x0800:
            vasResult=@"Approved";
            break;

        default:
            break;
    }

    [s appendFormat:@"Result: %@\n",vasResult];

    for (TLV *t in vas) {
        [s appendFormat:@"VAS: %@\n",t];
    }

    displayAlert(@"Apple VAS", s);
}

-(IBAction)onEMVTransaction:(id)sender
{
    NSError *error=nil;

    EMV_COMMAND(@"EMV Initialize",[dtdev emv2Initialise:&error]);

    [self loadConfig]; //don't fail if config loading fails, as lineas with vas for example does not support configuration

    shouldUpdate=false;

    [dtdev emv2SetPINOptions:PIN_ENTRY_AUTOMATIC error:nil];

    //enable apple VAS
    TLV *tagCD=[TLV tlvWithHexString:@"01" tag:0xCD];
    TLV *tagVAS=[TLV tlvWithData:[self getVASConfig] tag:0xE0];

    NSData *initData=[TLV encodeTags:@[tagCD, tagVAS]];

    //amount: $1.00, currency code: USD(840), according to ISO 4217
    EMV_COMMAND(@"EMV Init",[dtdev emv2SetTransactionType:0 amount:100 currencyCode:840 error:&error]);

//    uint8_t t[]={0xE0,0x81,0x94,0x9F,0x25,0x20,0x3F,0x22,0x54,0x3B,0xAF,0x0A,0xC5,0xE4,0xAB,0xFC,0x25,0x68,0x1A,0x6E,0xBF,0x6E,0xDF,0x5A,0xC1,0x96,0x74,0x6C,0x55,0xF4,0xD4,0x37,0x08,0x19,0xFF,0xF9,0x21,0xC3,0x9F,0x25,0x20,0xD6,0xDD,0x8C,0x2A,0xF4,0xB1,0x7A,0x2B,0xDB,0xF6,0x86,0x95,0xF8,0xFA,0x18,0x98,0x50,0xCF,0xB3,0x52,0x9C,0x84,0x3B,0xF7,0x20,0x7F,0xFE,0xBA,0x76,0x3D,0xAC,0x9A,0x9F,0x25,0x20,0xBA,0xD0,0xE2,0x1D,0x11,0x89,0x86,0x4F,0xEF,0x53,0xD2,0xEA,0xEC,0xD3,0x1D,0x1A,0xB6,0xE3,0x4C,0x95,0x6F,0xC2,0xAA,0xC2,0x99,0xD6,0x7A,0x5A,0xCB,0x6D,0xD8,0x42,0x9F,0x29,0x19,0x68,0x74,0x74,0x70,0x73,0x3A,0x2F,0x2F,0x70,0x61,0x73,0x73,0x6D,0x61,0x6E,0x2E,0x61,0x70,0x70,0x6C,0x65,0x2E,0x63,0x6F,0x6D,0x9F,0x22,0x02,0x01,0x00,0xDF,0xC9,0x01,0x01,0x00,0xDF,0xC9,0x02,0x01,0x02};
//    initData=[NSData dataWithBytes:t length:sizeof(t)];
    //start the transaction, transaction steps will be notified via emv2On... delegate methods
    EMV_COMMAND(@"EMV Start Transaction",[dtdev emv2StartTransactionOnInterface:EMV_INTERFACE_CONTACTLESS flags:0 initData:initData timeout:10 error:&error]);

    if(error)
    {
        [dtdev emv2Deinitialise:&error];
        [progressViewController.view removeFromSuperview];
    }else
        [self.view addSubview:[ProgressViewController progressWithMessage:@"Transaction starting"].view];
}

-(NSData *)getVASConfig
{
    NSMutableArray<TLV *> *vastlv=[NSMutableArray array];
    NSData *mid;

    mid=vasGetMID(tfMID1.text);
    if(mid)
        [vastlv addObject:[TLV tlvWithData:mid tag:0x9F25]];
    mid=vasGetMID(tfMID2.text);
    if(mid)
        [vastlv addObject:[TLV tlvWithData:mid tag:0x9F25]];
    mid=vasGetMID(tfMID3.text);
    if(mid)
        [vastlv addObject:[TLV tlvWithData:mid tag:0x9F25]];

    [vastlv addObject:[TLV tlvWithString:tfMURL.text tag:0x9F29]];

    [vastlv addObject:[TLV tlvWithInt:segProtocolMode.selectedSegmentIndex nBytes:1 tag:0xDFC901]];
    [vastlv addObject:[TLV tlvWithInt:segTerminalMode.selectedSegmentIndex nBytes:1 tag:0xDFC902]];

    NSData *vasdata=[TLV encodeTags:vastlv];

    return vasdata;
}

-(bool)loadConfig
{
    NSError *error;

    [self savePrefs];

    NSData *configuration=[Config paymentGetConfigurationFromXML:@"contactless.xml"];

    NSData *vasdata=[self getVASConfig];

    NSArray *cfgtlv=[TLV decodeTags:configuration];
    for (int i=0;i<cfgtlv.count;i++)
    {
        TLV *tag=[cfgtlv objectAtIndex:i];
        if(tag.tag==0xE0)
        {
            NSMutableData *d=[NSMutableData dataWithData:tag.data];
            [d appendData:vasdata];
            //add vas
            tag.data=d;
        }
    }
    configuration=[TLV encodeTags:cfgtlv];

    if(![dtdev emv2LoadContactlessConfiguration:configuration configurationIndex:0 error:&error])
    {
        //displayAlert(@"Error", [NSString stringWithFormat:@"Operation failed with error: %@",error.localizedDescription]);
        return false;
    }
    return true;
}

-(void)savePrefs
{
    NSUserDefaults *prefs=[NSUserDefaults standardUserDefaults];
    [prefs setValue:tfMID1.text forKey:@"MID1"];
    [prefs setValue:tfMID2.text forKey:@"MID2"];
    [prefs setValue:tfMID3.text forKey:@"MID3"];
    [prefs setValue:tfMURL.text forKey:@"MURL"];
    [prefs setInteger:segProtocolMode.selectedSegmentIndex forKey:@"ProtocolMode"];
    [prefs setInteger:segTerminalMode.selectedSegmentIndex forKey:@"TerminalMode"];
    [prefs synchronize];
}

-(void)viewWillDisappear:(BOOL)animated
{
    [super viewWillDisappear:animated];
    [self savePrefs];
}

- (void)viewDidLoad {
    [super viewDidLoad];
    // Do any additional setup after loading the view, typically from a nib.
    dtdev=[DTDevices sharedDevice];
    dtdev.delegate=self;
    [dtdev connect];

    NSUserDefaults *prefs=[NSUserDefaults standardUserDefaults];
    tfMID1.text=[prefs stringForKey:@"MID1"];
    tfMID2.text=[prefs stringForKey:@"MID2"];
    tfMID3.text=[prefs stringForKey:@"MID3"];
    tfMURL.text=[prefs stringForKey:@"MURL"];
    segProtocolMode.selectedSegmentIndex=[prefs integerForKey:@"ProtocolMode"];
    segTerminalMode.selectedSegmentIndex=[prefs integerForKey:@"TerminalMode"];
}


- (void)didReceiveMemoryWarning {
    [super didReceiveMemoryWarning];
    // Dispose of any resources that can be recreated.
}


@end
